import bpy
import struct
"""
Blender 2.59 File Selector
By Jacob Valenta

I write this because I couldnt easily find a way to create a window
in Blender where I can select a file and then use that file. This
creates an easy way to open a file, and it works out of the box;
copy and paste the code into Blender and run it. As a side note, after
searching for quite a while, this code was found on the Blender
documentation, tucked away in its far corners.
"""
class CustomDrawOperator(bpy.types.Operator):
    bl_idname = "object.custom_draw"
    bl_label = "Import"
 
    filepath = bpy.props.StringProperty(subtype="FILE_PATH")
 
    my_float = bpy.props.FloatProperty(name="Float")
    my_bool = bpy.props.BoolProperty(name="Toggle Option")
    my_string = bpy.props.StringProperty(name="String Value")
 
    def execute(self, context):
        print('Execute Function')
        print('filepath = ',self.filepath)
        return {'FINISHED'}
 
    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
 
    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.label(text="Custom Interface!")
 
        row = col.row()
        row.prop(self, "my_float")
        row.prop(self, "my_bool")
 
        col.prop(self, "my_string")
 
bpy.utils.register_class(CustomDrawOperator)
 
# test call
ret_val = bpy.ops.object.custom_draw('INVOKE_DEFAULT')
"""
This code is very minimal, and could be condensed even more. What it will
do is bring up a file selection window. You are then able to browse though
folders and select a file and then you can press the button once you get
the file you want.

This script doesnt actually do anything with that file, but what ever
you want the script to do with have to be coded in the execute() function.

There are three variables set up that are later referenced in the draw()
function. These variables help create the menu on the side of the window.
With this, you can create options for the user to provide and affect what
the script does with the file.

The draw() function places these objects on rows and columns. Every time
you use the col.prop() function, the next time you use it, that value
will be placed on the next column. (actually, these names are backwards)
and every time you use the row.prop() it will add another variable to that
same row.

To access the filename to actually open the file with python, it would be,
in this example, self.filename, and likewise, to access the other variables,
you would use self.my_float

Well that is about it for basics on creating a file selector with
Blender 2.59
"""
